
; Co-cartesian closed category
;   i.e. disjoint unions implemented via CPS conversion

(define o 'o)				;answer type

(define (cont a) (-> (-> a o) o))

(define (=> a b) (-> a (cont b)))        ;Arrow in kleisli category

(define (% a b) (-> (& (-> a o) (-> b o)) o)) ;Implementation of sum

(define (<- b c) (& c (-> b o)))	      ;Coexponential

(define (unique a)
  (let ((stream (pr a)))
    (if (empty-stream? stream)
	(error "no proof" a)
	(if (empty-stream? (tail stream))
	    (head stream)
	    (error "proof not unique" a)))))

(define ccompose
  (unique (-> (& (=> q r) (=> p q)) (=> p r))))

(define (check= a b)
  (if (equal? (normal a) (normal b))
      (begin (write 'ok) (newline))
      (begin (write `(loser ,(normal a) ,(normal b)))
	     (newline))))


; Verify that the combinators found by the theorem prover make the CPS
; monad's Kleisli category a co-CCC.

(define (verify-co-ccc)
  (let ((abort   (unique (=> o p)))
	(i1      (unique (=> p (% p q))))
	(i2      (unique (=> q (% p q))))
	(casef   (unique (-> (& (=> p r) (=> q r))
			     (=> (% p q) r))))
	(cocurry (unique (-> (=> r (% p q))
			     (=> (<- q r) p))))
	(pa      (unique (=> r (% (<- q r) q))))
	(id      (unique (=> p p)))
	(o       (lambda (f g)
		   `(,ccompose (pair ,f ,g))))
	(f 'f)
	(g 'g)
	(h 'h))

    (check= (o `(,casef (pair ,f ,g)) i1) f)
    (check= (o `(,casef (pair ,f ,g)) i2) g)

    ;; f = {f_*, id} o pa = [i1 o f_*, i2] o pa
    (let ((consumer `(,casef (pair ,g ,h)))
	  (try (lambda (f)
		 (o `(,casef (pair (,ccompose (pair ,i1 (,cocurry ,f))) ,i2))
		    pa))))
      ;; (check= (try f) f) loses, but the following weaker results hold.
      (check= (o consumer (try f))
	      (o consumer f))
      (check= (o i1 f) (try (o i1 f)))
      (check= (o i2 f) (try (o i2 f))))))

;(define pair cons)
;(define left car)
;(define right cdr)
;
;(define (reflect name type)
;  (eval `(let ((,name ,(unique type))) ,name) (interaction-environment)))
;
;(define i1 (reflect 'i1 (=> p (% p q))))
;(define i2 (reflect 'i2 (=> q (% p q))))
;(define casef   (reflect 'casef (-> (& (=> p r) (=> q r))
;                                    (=> (% p q) r))))
;(define cocurry (reflect 'cocurry (-> (=> r (% p q))
;                                      (=> (<- q r) p))))
;(define pa      (reflect 'pa (=> r (% (<- q r) q))))
;(define id      (reflect 'id (=> p p)))
;(define ccomp   (reflect 'ccomp (-> (& (=> q r) (=> p q)) (=> p r))))
;
;(define (o g f) (ccomp (pair g f)))
;(define (done x)
;  (if (procedure? x)
;      (x (pair (lambda (y) `(i1 ,y)) (lambda (z) `(i2 ,z))))
;      x))
;(define (run foo) (foo done))
;
;(define (f x) (lambda (k) (k `(f ,x))))
;(define (g x) (lambda (k) (k `(g ,x))))
