
; Reduce to normal form.

; This is kind of sloppy, because variable names aren't canonicalized, so
; some (pair (left x) (right x)) reductions are missed.  E.g.
;  (pair (left (f (lambda (x) x)))
;        (right (lambda (z) ((f (lambda (y) y)) z))))
; should reduce to (f (lambda (x0) x0)), but doesn't.

; The occurs? check and beta-reduction rules don't appear to be
; correct, but I think this doesn't matter, since there shouldn't be
; any opportunities for capture in the first place.

(define (normal x)
  (set! *counter* 0)
  (let ((x (normalize x '())))
    ;; 2nd pass canonicalizes variable names.
    (set! *counter* 0)
    (normalize x '())))

; Assumes that term has no free variables of the form xNN where NN is
; an integer.

(define (normalize term env)
  (cond ((variable? term)
	 (let ((probe (assq term env)))
	   (if probe (cdr probe) term)))
	((and-intro? term)
	 (let ((left (normalize (left-operand term) env))
	       (right (normalize (right-operand term) env)))
	   (if (and (and-elim-left? left)
		    (and-elim-right? right)
		    (equal? (operand left) (operand right)))
	       (operand left)		;Surjective pairing
	       `(pair ,left ,right))))
	((and-elim-left? term)
	 (let ((arg (normalize (operand term) env)))
	   (if (and-intro? arg)
	       (left-operand arg)
	       `(left ,arg))))
	((and-elim-right? term)
	 (let ((arg (normalize (operand term) env)))
	   (if (and-intro? arg)
	       (right-operand arg)
	       `(right ,arg))))
	((or-intro-left? term)
	 `(in-left ,(normalize (operand term) env)))
	((or-intro-right? term)
	 `(in-right ,(normalize (operand term) env)))
	((or-elim? term)
	 (let ((arg (normalize (cadr term) env)))
	   (cond ((or-intro-left? arg)
		  (normalize `(,(caddr term) ,(operand arg)) env))
		 ((or-intro-right? arg)
		  (normalize `(,(cadddr term) ,(operand arg)) env))
		 (else
		  (let ((left-fun (normalize (caddr term) env))
			(right-fun (normalize (cadddr term) env)))
		    (if (and (eq? left-fun 'in-left)
			     (eq? right-fun 'in-right))
			arg
			`(sum-case ,arg ,left-fun ,right-fun)))))))
	((imp-intro? term)
	 (let* ((var (generate-variable 'x))
		(body (normalize (lambda-body term)
				 (cons (declare (lambda-variable term) var)
				       env))))
	   (if (and (imp-elim? body)
		    (eq? (operand body) var)
		    (not (occurs? var (operator body))))
	       (operator body)		;Eta reduce
	       `(lambda (,var) ,body))))
	((imp-elim? term)
	 (let ((fun (normalize (operator term) env))
	       (arg (normalize (operand term) env)))
	   (if (imp-intro? fun)
	       (normalize (lambda-body fun)
			  (cons (declare (lambda-variable fun) arg)
				env))
	       `(,fun ,arg))))
	(else term)))


; Terms

(define variable? symbol?)

(define (operator-is name)
  (lambda (form) (and (pair? form) (eq? (car form) name))))

(define and-intro? (operator-is 'pair))
(define and-elim-left? (operator-is 'left))
(define and-elim-right? (operator-is 'right))

(define or-intro-left? (operator-is 'in-left))
(define or-intro-right? (operator-is 'in-right))
(define or-elim? (operator-is 'sum-case))

(define imp-intro? (operator-is 'lambda))
(define lambda-variable cadar)
(define lambda-body right-operand)
(define (imp-elim? form)    ;wrong
  (and (pair? form) (not (null? (cdr form))) (null? (cddr form))))

(define operator car)
(define operand cadr)

(define (occurs? var term)
  (or (eq? var term)
      (and (pair? term)
	   (some (lambda (subterm)
		   (occurs? var subterm))
		 term))))


(define declare cons)

