


; prove : proposition -> assumption-set -> answer

(define (prove prop)
  (cond ((conj? prop)
	 (and-intro (prove (left-operand prop))
		    (prove (right-operand prop))))
	((disj? prop)
	 (or-intro (prove (left-operand prop))
		   (prove (right-operand prop))))
	((impl? prop)
	 (imp-intro (suppose (left-operand prop))
		    (prove (right-operand prop))))
	(else
	 (backward-chain prop))))

; suppose : proposition -> assumption-set -> assumption-set

(define (suppose prop)
  (cond ((conj? prop)
	 (and-elim (suppose (left-operand prop))
		   (suppose (right-operand prop))))
	((disj? prop)
	 (or-elim (suppose (left-operand prop))
		  (suppose (right-operand prop))))
	((impl? prop)
	 (imp-elim (prove (left-operand prop))
		   (suppose (right-operand prop))))
	(else
	 (forward-chain prop))))


; Introduction rules (prove)

(define (and-intro y1 y2)
  (lambda (Gamma)
    (product-stream (lambda (term1 term2)
		      `(pair ,term1 ,term2))
		    (y1 Gamma)
		    (y2 Gamma))))

(define (or-intro y1 y2)
  (lambda (Gamma)
    (merged-stream (map-stream (lambda (term1)
				 `(in-left ,term1))
			       (y1 Gamma))
		   (map-stream (lambda (term2)
				 `(in-right ,term2))
			       (y2 Gamma)))))

(define (imp-intro z y)
  (lambda (Gamma)
    (let ((var (generate-variable 'x)))
      (map-stream (lambda (term)
		    `(lambda (,var) ,term))
		  (z var Gamma y)))))

(define (backward-chain prop)		;Prove atomic proposition
  (lambda (Gamma)
    (try-rules Gamma
	       (lambda (Gamma)
		 (list->stream
		  (map cdr
		       (filter (lambda (prop+term)
				 (equal? prop (car prop+term)))
			       (assumption-set-atomics Gamma))))))))


; Elimination rules (suppose)

(define (and-elim z1 z2)
  (lambda (term Gamma y)
    (z1 `(left ,term)
	Gamma			;Assume `(left ,term)
	(lambda (new-Gamma)
	  (z2 `(right ,term) new-Gamma y))))) ;Assume right & prove y

(define (or-elim z1 z2)
  (lambda (term Gamma y)
    (let ((foo (generate-variable 'l))
	  (bar (generate-variable 'r)))
      (product-stream (lambda (term1 term2)
			`(sum-case ,term
			   (lambda (,foo) ,term1)
			   (lambda (,bar) ,term2)))
		      (z1 foo Gamma y) 	;Prove y having assumed left
		      (z2 bar Gamma y)))))  ;Prove y having assumed right

(define (imp-elim y z)
  (lambda (fun Gamma y2)		;fun has type A -> B
    (y2 (make-assumption-set (assumption-set-atomics Gamma)
			     (cons (make-rule y z fun)
				   (assumption-set-rules Gamma))))))

(define (forward-chain prop)	;Suppose atomic proposition
  (lambda (term Gamma y)
    (y (make-assumption-set (cons (cons prop term)
				  (assumption-set-atomics Gamma))
			    (assumption-set-rules Gamma)))))

; Try to find a rule that now applies.

(define (try-rules Gamma y)
  (let ((answer (merge-streams (map (lambda (rule)
				      (lambda () (try-rule rule Gamma y)))
				    (assumption-set-rules Gamma)))))
    (if (empty-stream? answer)
	(y Gamma)
	(without-duplicates answer))))

(define (try-rule rule Gamma y)
  (let ((Gamma-without-rule
	 (make-assumption-set (assumption-set-atomics Gamma)
			      (remove rule (assumption-set-rules Gamma)))))
    (merge-stream-of-streams
       (map-stream (lambda (arg)
		     ((rule-supposer rule)
		      `(,(rule-function rule) ,arg)
		      Gamma-without-rule
		      (lambda (new-Gamma)
			(try-rules new-Gamma y))))
		   ((rule-prover rule) Gamma-without-rule)))))


; Assumption sets

(define (make-assumption-set atomics rules)
  (cons atomics rules))
(define assumption-set-atomics car)
(define assumption-set-rules cdr)

(define (filter pred lst)
  (reduce (lambda (x rest)
	    (if (pred x) (cons x rest) rest))
	  '()
	  lst))
(define (reduce cons nil list)		;used by length, append, etc.
  (if (null? list)
      nil
      (cons (car list) (reduce cons nil (cdr list)))))

; Rules

(define (make-rule prover supposer term) (list prover supposer term))
(define rule-prover car)
(define rule-supposer cadr)
(define rule-function caddr)

(define (remove rule rule-set)
  (if (null? rule-set)
      (error "shouldn't happen" `(remove ,rule ,rule-set))
      (if (same-rule? rule (car rule-set))
	  (cdr rule-set)
	  (cons (car rule-set) (remove rule (cdr rule-set))))))

(define same-rule? eq?)


; Variable generation

(define *counter* 0)

(define (generate-variable prefix)
  (concatenate-symbol prefix
		      (let ((n *counter*))
			(set! *counter* (+ *counter* 1))
			n)))
		      

(define (concatenate-symbol . things)
  (string->symbol
    (apply string-append
	   (map (lambda (thing)
		  (cond ((string? thing) thing)
			((symbol? thing) (symbol->string thing))
			((number? thing) (number->string thing '(heur)))
			(else (error "weird argument to concatenate-symbol"
				     thing))))
		things))))

(define (pr prop)
  (set! *counter* 0)
  ((prove prop) usual-assumption-set))

(define usual-assumption-set
  (make-assumption-set (list (cons 1 'unit)) '()))

; Propositions

(define (conj? prop) (and (pair? prop) (eq? (car prop) '&)))
(define (disj? prop) (and (pair? prop) (eq? (car prop) '^)))
(define (impl? prop) (and (pair? prop) (eq? (car prop) '->)))

(define (left-operand prop) (cadr prop))
(define (right-operand prop) (caddr prop))


(define *s* #f)
(define (test p)
  (write p)
  (display " =>")
  (newline)
  (set! *s* (pr p))
  (write-stream *s*)
  (newline))
